--@author      : Yang Min
--@date        : 2021-02-20
--@description : lua script for face reshape with autotoucher function
--@version     : 1.0

local apolloengine  = require "apolloengine"
local apollonode    = require "apolloutility.apollonode"
local cv            = require "computervisionfunction"
local autotoucher   = require "autotoucherfunction"
local venusjson     = require "venusjson"
local stretchconfig = require "atfacereshape.stretch_config"

local FaceReshape = {
    versionnum = -1,
    config = {},
    handleIds = {},
    deformIds = {},
    deformDeltas = {},
    stretchCoefs = {}
};

function FaceReshape:Create(layer_sequence, scene, reshaper)
    ERROR("[FaceReshape]: Create");
    -- if reshaper is nil, use autotoucher as default
    if (reshaper == nil) then
        self.reshaper = autotoucher.AutoToucherManager();
    else
        self.reshaper = reshaper;
    end

    local res = self.reshaper:InitRenderNode(scene, layer_sequence);

    if res > 0 then
        ERROR("[FaceReshape] InitRenderNode return value: "..res);
        self.initialized = false;
    end

    -- assets/resource/facelift/params_stretch.json
    local params = venusjson.LaodJsonFile("docs:facelift/params_stretch.json");
    ERROR("set config:");
    for k,v in pairs(params) do
        ERROR("config "..v["Name"]);
        self.config[v["Name"]] = v;
    end
    self.initialized = false;

    self.CaptureInput = apolloengine.TextureEntity();
    self.CaptureInput:PushMetadata(apolloengine.TextureReferenceMetadata(apolloengine.DeviceResource.DEVICE_CAPTURE));
    self.CaptureInput:CreateResource();

    self.virtual_node   = apollonode.VirtualNode();
    local recognition   = self.virtual_node:CreateComponent(apolloengine.Node.CT_CV_RECOGNITION);
    recognition.Enable  = true;
    recognition.Mode    = cv.IVisionComponent.VIDEO;
    recognition.Type    = cv.RecognitionComponent.cvFace + cv.RecognitionComponent.cvForeheadDetection;
    recognition.ShareCV = true;
    recognition:SetTexture(self.CaptureInput);
    self.recognition    = recognition;

    self.preUpdateResult = 0;
    self.updateResult    = 0;
    self.initialized     = true;
    ERROR("[FaceReshape]: Create done!");
end

function FaceReshape:SetShow(isShow)
    self.reshaper:SetShow(isShow);
end

function FaceReshape:PreUpdate(def)

    if not self.initialized then
        ERROR("[FaceReshape] not initialized");
        return false;
    end

    if _KRATOSEDITOR then

    else
        self.virtual_node.node:Update(def);
    end

    local res = self.reshaper:PreUpdateRenderNode(self.recognition);

    if res ~= 0 then
        if self.preUpdateResult ~= res then
            ERROR("[FaceReshape] PreUpdate return value: "..res);
        end
    end
    self.preUpdateResult = res;

    return (res == 0);
end

function FaceReshape:Update(texture, sequence)
    if not self.initialized then
        ERROR("[FaceReshape] not initialized");
        return false;
    end

    local res = self.reshaper:UpdateRenderNode(self.recognition, texture, sequence);

    if res ~= 0 then
        if self.updateResult ~= res then
            ERROR("[FaceReshape] Update return value: "..res);
        end
    end
    self.updateResult = res;
end

function FaceReshape:SetFaceReshapeStrengthWithType(type, strength)
    ERROR("[FaceReshape] SetFaceReshapeStrengthWithType "..type..", "..strength);
    local stretchName = stretchconfig[type];
    if stretchName==nil then
        if self.reshaper ~= nil then
            self.reshaper:SetFaceReshapeStrengthWithType(math.ceil(strength), type);
        end
    else
        self.stretchCoefs[stretchName] = strength/100;
        local handleIds = {};
        local deformIds = {};
        local handleDeltas = {};
        for i=1, 465 do
            table.insert(handleDeltas, 0);
            table.insert(handleDeltas, 0);
        end
        for k,v in pairs(self.stretchCoefs) do
            local stretchData = self.config[k];
            local offsets = v >= 0 and stretchData.PosDelta or stretchData.NegDelta;
            local coef = v >= 0 and v or -v;
            for i=1,#stretchData.Handles do
                handleIds[stretchData.Handles[i]] = true;
            end
            for i=1,#stretchData.DeformIds do
                deformIds[stretchData.DeformIds[i]] = true;
            end
            for i=1,#stretchData.Handles do
                local index = stretchData.Handles[i];
                handleDeltas[2*index+1] = handleDeltas[2*index+1] + offsets[2*i-1]*coef;
                handleDeltas[2*index+2] = handleDeltas[2*index+2] + offsets[2*i]*coef;
            end
        end
        local newHandleIds = {};
        local newDeformIds = {};
        local newHandleDeltas = {};
        for k,v in pairs(handleIds) do
            table.insert(newHandleIds, k);
        end
        for k,v in pairs(deformIds) do
            table.insert(newDeformIds, k);
        end
        -- computed normalized factor for delta in standardface template
        -- (normalized delta value)
        local domainX_x = 15.649832; -- x component of face X axis
        local domainX_y = -0.260100; -- y component of face X axis
        local domainY_x = 0.472007; -- x component of face Y axis
        local domainY_y = 27.148807; -- y component of face Y axis
        for k,v in pairs(handleIds) do
            table.insert(newHandleDeltas, handleDeltas[2*k+1] * domainX_x + handleDeltas[2*k+2] * domainX_y);
            table.insert(newHandleDeltas, handleDeltas[2*k+1] * domainY_x + handleDeltas[2*k+2] * domainY_y);
        end
        self.handleIds = newHandleIds;
        self.deformIds = newDeformIds;
        self.handleDeltas = newHandleDeltas;
        self.reshaper:SetStretchParams(self.handleIds, self.deformIds, self.handleDeltas);
    end
    return true;
end

function FaceReshape:Destroy()

    if self.virtual_node ~= nil then
        self.virtual_node:Destroy();
        self.virtual_node = nil;
    end

    if self.reshaper ~= nil then
        self.reshaper:ReleaseRenderNode();
        self.reshaper = nil;
    end

    self.preUpdateResult = 0;
    self.updateResult    = 0;
    self.initialized = false;
end

return FaceReshape;