local venuscore = require "venuscore"
local venusjson = require "venusjson"
local apolloengine = require "apolloengine"
local mathfunction = require "mathfunction"
local libmasque = require "libmasque"
local quatfilter = require "math.lowpassfilter.oneeurofilter_quat"
local Hub = require "behavior.vtuber_behavior.vtuber_msg_hub"
local likeapp = require "likeapp"

local EyeRotate = venuscore.VenusBehavior:extend("EyeRotate")

local LOOK_FROM = mathfunction.vector3(0.0, 0.0, 1.0)

function EyeRotate:new()
  self.NodesConfig = ""
  self.eye_trans = {}
  self.Sending = false
  self.Receiving = false
  self.RotateAxis = 1;
end

function EyeRotate:_OnAwake()
  if not self.Tracking then
    if self.Node:HaveComponent(apolloengine.Node.CT_TRACK) then
      self.Tracking = self.Node:GetComponent(apolloengine.Node.CT_TRACK);
    end
  end
  self:LoadConfig()
  self:GetEyeNodes()
end

function EyeRotate:_OnStart()
  self.start_time = venuscore.ITimerSystem:GetTimevalue();
end

local function matrix_to_quaternion(mat) 
  if mat == nil or (#mat < 16) then
    return mathfunction.Quaternion(0, 0, 0, 1)
  end
  local tr = mat[1] + mat[6] + mat[11]
  local w = math.sqrt(tr + 1) * 0.5
  if w < 0.0001 then
    return mathfunction.Quaternion(0, 0, 0, 1)
  end
  local inv_4w = 0.25 / w
  local x = (mat[10] - mat[7]) * inv_4w
  local y = (mat[3] - mat[9]) * inv_4w
  local z = (mat[5] - mat[2]) * inv_4w
  local q = mathfunction.Quaternion(-x, -y, -z, w)
  return q
end

function EyeRotate:_OnLateUpdate(delta)
  if not self.Tracking and not self.Receiving then
    return
  end

  if self.Receiving then

    local data = Hub:Get(Hub.ReceivingId)
    local rotLEye = data.rotLEye
    local rotREye = data.rotREye
    if rotLEye and self.eye_trans[1] then
      self.eye_trans[1].component:SetLocalRotation(rotLEye)
    end
    if rotREye and self.eye_trans[2] then
      self.eye_trans[2].component:SetLocalRotation(rotREye)
    end

    return
  end

  local use_arkit = false

  if _KRATOSEDITOR then
  else
    use_arkit = likeapp.AI:GetARKitFlag()
  end
  
  if use_arkit then
    local faceNum = likeapp.AI:GetARKitFaceNum()
    local rotLEye = mathfunction.Quaternion()
    local rotREye = mathfunction.Quaternion()

    if faceNum > 0 then
      local leftEyeRot = likeapp.AI:GetARKitLeftEyeTransform()
      local rightEyeRot = likeapp.AI:GetARKitRightEyeTransform()

      rotLEye = matrix_to_quaternion(leftEyeRot)
      rotREye = matrix_to_quaternion(rightEyeRot)
      
      if self.RotateAxis == 0 then
        local eulerL = rotLEye:ToEulerAngle();
        local convert_quatL = mathfunction.Quaternion();
        convert_quatL:RotateXYZ(eulerL:x(), eulerL:z(), eulerL:y());
        rotLEye = convert_quatL;

        local eulerR = rotREye:ToEulerAngle();
        local convert_quatR = mathfunction.Quaternion();
        convert_quatR:RotateXYZ(eulerR:x(), eulerR:z(), eulerR:y());
        rotREye = convert_quatR;
      end
    end

    self.eye_trans[1].component:SetLocalRotation(rotLEye)
    self.eye_trans[2].component:SetLocalRotation(rotREye)

    if self.Sending then
      Hub:PutEyeRotation(Hub.SendingId, rotLEye, rotREye)
    end

    return
  end

  local lookat = mathfunction.vector3(0.0, 0.0, 1.0)
  if self.Tracking:IsTracked() then
    lookat = self.Tracking:GetLookat()    
    -- set eye rotate threshold in y-axis, to prevent rolling eye issue
    lookat = mathfunction.vector3(lookat:x(), math.clamp((lookat:y() + self.y_offset) * self.y_factor, -self.y_threshold, self.y_threshold), lookat:z())
  end

  if lookat == nil then
    LOG("[Eye rotate behavior] look at is invalid!")
    return
  end

  local quat = mathfunction.Mathutility:AxisToAxis(LOOK_FROM, lookat)

  -- low pass filter, to reduce shake of eyeball
  if self.quatFilter == nil then    
    self.quatFilter = quatfilter(
      venuscore.ITimerSystem:GetTimevalue() - self.start_time, 
      quat,
      2
    );
  else
    quat = self.quatFilter:filter(
      venuscore.ITimerSystem:GetTimevalue() - self.start_time, 
      quat
    );
  end

  if self.RotateAxis == 0 then
    local  euler = quat:ToEulerAngle();
    local convert_quat = mathfunction.Quaternion();
    convert_quat:RotateXYZ(euler:x(), euler:z(), euler:y());
    quat = convert_quat;
  end

  local rotations = {}
  local cos_theta = LOOK_FROM:Dot(lookat)
  for _, trans in ipairs(self.eye_trans) do
    if not _KRATOSEDITOR or not venuscore.isNil(trans.node) then
      local rot;
      if cos_theta >= self.angle_threshold then -- eye ball rotation angle is valid, degree from 0 to arccos(threshold)
        rot = trans.start * quat
        trans.component:SetLocalRotation(rot)
      else
        rot = trans.component:GetLocalRotation()
      end

      table.insert(rotations, rot)
    end
  end
  if self.Sending then
    Hub:PutEyeRotation(Hub.SendingId, rotations[1], rotations[2])
  end
end

function EyeRotate:LoadConfig()
  if self.NodesConfig == "" then
    return
  end
  local file = venusjson.LaodJsonFile(self.NodesConfig);
  if file ~=nil then
    self.node_names = file.nodes
    self.y_offset = file.y_offset or 0.0
    self.y_threshold = file.y_threshold or 0.268 -- tan15
    self.y_factor = file.y_factor or 1
    self.angle_threshold = file.angle_threshold or 0.259 -- cos75
  end
end

function EyeRotate:GetEyeNodes()
  if not self.Node or not self.node_names then
    return
  end
  self.eye_trans = {}
  for _, name in ipairs(self.node_names) do
    local eye_node = self.Node:GetNodeByName(name)
    if eye_node and eye_node:HaveComponent(apolloengine.Node.CT_TRANSFORM) then
        local trans_component = eye_node:GetComponent(apolloengine.Node.CT_TRANSFORM)
        table.insert(self.eye_trans, {
          node = eye_node,
          component = trans_component,
          start = trans_component:GetLocalRotation()
        })
    end
  end
end

EyeRotate:MemberRegister(
  "Tracking",
  venuscore.ScriptTypes.ReferenceType(
    libmasque.TrackingComponent:RTTI()
  )
)

EyeRotate:MemberRegister("NodesConfig",  
  venuscore.ScriptTypes.FilePathType(
    {"json"},
    nil,
    function(thiz, value)
        thiz.NodesConfig = value
        thiz:LoadConfig()
        thiz:GetEyeNodes()
    end
))

EyeRotate:MemberRegister("RotateAxis");
EyeRotate:MemberRegister("Sending");
EyeRotate:MemberRegister("Receiving");

return EyeRotate
