--@author      : Jin Shiyu
--@date        : 2020-11-02
--@description : lua script for face beauty
--@version     : 1.0

local apolloengine  = require "apolloengine"
local apollonode    = require "apolloutility.apollonode"
local mathfunction  = require "mathfunction"
local venuscore     = require "venuscore"
local quad_render   = require "facebeautyutils.face_beauty_quad_render"
local cv            = require "computervisionfunction"
local autotoucher   = require "autotoucherfunction"

--[[
    Face beauty on one frame
    NOTE:
    It can be used in testshell main and editor behavior which should offer one main camera.
]]

local FaceBeauty = {};

function FaceBeauty:Create(layer_sequence, scene)
    if not APP_IMO_VENUS then
    self.hightlight_texture_path    = "docs:facebeauty/resource/shadow_light_mask511.png"
    self.mask_texture_path          = "docs:facebeauty/resource/standard_face_v4.png"
    self.teeth_lookup_texture_path  = "docs:facebeauty/resource/whiten_teeth_lut.png"
    self.eye_mask_texture_path      = "docs:facebeauty/resource/eye_face_mask.png"
    end  -- APP_IMO_VENUS
    self.clear_lookup_texture_path  = "docs:facebeauty/resource/clear_lut.png"
    self.whiten_lookup_texture_path = "docs:facebeauty/resource/whiten_lut.png"

    --[[
        0: current strength of highlight (0-1)
        1: current strength of brighten eye (0-1)
        2: current strength of nasolabial removal (0-1)
        3: current strength of eyebag removal (0-1)
        4: current strength of whiten teeth (0-1)
        5: current strength of clear (0-1)
        6: current strength of whiten (0-1)
    ]]
    self.beauty_strengths = { [0]=0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
    self.beauty_typeid = {HIGHLIGHT=0, BRIGHT_EYE=1, NASOLABIAL=2, EYEBAG=3, WHITEN_TEETH=4, CLEAR=5, WHITEN=6};

    self.scaled_factor    = 0.5;            -- scaled factor of downsample
    self.offset_factor    = 2.8958829;      -- factor of texture sampling step

    self.max_face_num     = 4;

    self.virtual_node     = nil;

    -- if self.is_show is false or self.need_clear_or_whiten is false, clear and whiten not render.
    self.need_clear_or_whiten        = false; -- will be set true if clear strength or whiten strength > 0

    -- if self.is_show is false or no face detected, autotoucher facebeauty items not render.
    self.need_autotoucher_facebeauty = false;

    -- if self.initialized is false or self.is_show is false, will deactivate the layer (need to deactivate the added rendernodes of the layer).
    self.is_show          = true;
    self.initialized      = false;
    self.preUpdateResult  = 0;
    self.updateResult     = 0;
    self.autotoucher = autotoucher.AutoToucherFaceBeauty();
    self:Initialize(layer_sequence, scene);
end

function FaceBeauty:UpdateTextureSize(frame_width, frame_height)

  local scaled_width  = frame_width * self.scaled_factor;
  local scaled_height = frame_height * self.scaled_factor;

  local scaler_updated     = self.image_scaler:Update(scaled_width, scaled_height);
  local vertical_updated   = self.vertical_box_blur:Update(scaled_width, scaled_height);
  local horizontal_updated = self.horizontal_box_blur:Update(scaled_width, scaled_height);

  -- update
  if scaler_updated and vertical_updated and horizontal_updated then
      self.vertical_box_blur:SetParameter("UNIFORM_INPUT_TEXTURE", self.image_scaler:GetResultImage());
      self.vertical_box_blur:SetParameter("UNIFORM_TEXEL_WIDTH_OFFSET", mathfunction.vector1(0));
      self.vertical_box_blur:SetParameter("UNIFORM_TEXEL_HEIGHT_OFFSET", mathfunction.vector1(self.offset_factor / scaled_height));

      self.horizontal_box_blur:SetParameter("UNIFORM_INPUT_TEXTURE", self.vertical_box_blur:GetResultImage());
      self.horizontal_box_blur:SetParameter("UNIFORM_TEXEL_WIDTH_OFFSET", mathfunction.vector1(self.offset_factor / scaled_width));
      self.horizontal_box_blur:SetParameter("UNIFORM_TEXEL_HEIGHT_OFFSET", mathfunction.vector1(0));
  end

end

function FaceBeauty:LoadTexture(texture_path)
    local texture = apolloengine.TextureEntity();
    texture:PushMetadata(apolloengine.TextureFileMetadata(
                                   apolloengine.TextureEntity.TU_STATIC,
                                   apolloengine.TextureEntity.PF_AUTO,
                                   1, false,
                                   apolloengine.TextureEntity.TW_REPEAT,
                                   apolloengine.TextureEntity.TW_REPEAT,
                                   apolloengine.TextureEntity.TF_LINEAR,
                                   apolloengine.TextureEntity.TF_LINEAR,
                                   venuscore.IFileSystem:PathAssembly(texture_path)));
    texture:SetJobType(venuscore.IJob.JT_SYNCHRONOUS);
    texture:CreateResource();
    return texture;
end

function FaceBeauty:InitQuadNode(seq, material_path)
    local render = apollonode.QuadNode();
    render:CreateResource(material_path, true);
    render:SetSequence(seq);
    render:GetNode():GetComponent(apolloengine.Node.CT_RENDER):SetRenderOrder(0);
    render:SetCull(false);
    render:SetShow(false);
    render:SetName("identity render");
    -- render:GetNode().Active = false;

    return render;
end

function FaceBeauty:Initialize(layer_sequence, scene)

    -- load texture
    if not APP_IMO_VENUS then
    self.mask_texture          = self:LoadTexture(self.mask_texture_path);
    self.teeth_lookup_texture  = self:LoadTexture(self.teeth_lookup_texture_path);
    self.eye_mask_texture      = self:LoadTexture(self.eye_mask_texture_path);
    self.hightlight_texture    = self:LoadTexture(self.hightlight_texture_path);
    end -- APP_IMO_VENUS
    self.clear_lookup_texture  = self:LoadTexture(self.clear_lookup_texture_path);
    self.whiten_lookup_texture = self:LoadTexture(self.whiten_lookup_texture_path);


    -- setup quad render
    local initial_width  = 360;
    local initial_height = 640;
    local max_sequence   = layer_sequence - self.max_face_num;

    self.clear_whiten        = quad_render("clear_whiten", max_sequence - 4, "clear_whiten_layer", "docs:facebeauty/material/clear_whiten.material", initial_width, initial_height);
    if not APP_IMO_VENUS then
    self.image_scaler        = quad_render("face_beauty_identity", max_sequence - 3, "identity_layer", "docs:facebeauty/material/identity.material", initial_width, initial_height);
    self.vertical_box_blur   = quad_render("face_beauty_vertical", max_sequence - 2, "vertical_blur_layer", "docs:facebeauty/material/box_blur.material", initial_width, initial_height);
    self.horizontal_box_blur = quad_render("face_beauty_horizontal", max_sequence - 1, "horizontal_blur_layer", "docs:facebeauty/material/box_blur.material", initial_width, initial_height);
    end

    self.identity_render = self:InitQuadNode(layer_sequence, "docs:facebeauty/material/identity.material");

    self.clear_whiten:SetParameter("UNIFORM_CLEAR_LUT_TEXTURE", self.clear_lookup_texture);
    self.clear_whiten:SetParameter("UNIFORM_WHITEN_LUT_TEXTURE", self.whiten_lookup_texture);

    if not APP_IMO_VENUS then
    self.vertical_box_blur:SetParameter("UNIFORM_INPUT_TEXTURE", self.image_scaler:GetResultImage());
    self.vertical_box_blur:SetParameter("UNIFORM_TEXEL_WIDTH_OFFSET", mathfunction.vector1(0));
    self.vertical_box_blur:SetParameter("UNIFORM_TEXEL_HEIGHT_OFFSET", mathfunction.vector1(self.offset_factor / initial_height));

    self.horizontal_box_blur:SetParameter("UNIFORM_INPUT_TEXTURE", self.vertical_box_blur:GetResultImage());
    self.horizontal_box_blur:SetParameter("UNIFORM_TEXEL_WIDTH_OFFSET", mathfunction.vector1(self.offset_factor / initial_width));
    self.horizontal_box_blur:SetParameter("UNIFORM_TEXEL_HEIGHT_OFFSET", mathfunction.vector1(0));

    self.CaptureInput = apolloengine.TextureEntity();
    self.CaptureInput:PushMetadata(apolloengine.TextureReferenceMetadata(apolloengine.DeviceResource.DEVICE_CAPTURE));
    self.CaptureInput:CreateResource();

    self.virtual_node       = apollonode.VirtualNode();
    local classify          = self.virtual_node:CreateComponent(apolloengine.Node.CT_CV_CLASSIFY);
    classify.Enable         = true;
    classify.Mode           = cv.IVisionComponent.VIDEO;
    classify.FaceExpression = true;
    classify:SetTexture(self.CaptureInput);
    self.classify = classify;

    local recognition   = self.virtual_node:CreateComponent(apolloengine.Node.CT_CV_RECOGNITION);
    recognition.Enable  = true;
    recognition.Mode    = cv.IVisionComponent.VIDEO;
    recognition.Type    = cv.RecognitionComponent.cvFace + cv.RecognitionComponent.cvForeheadDetection;
    recognition.ShareCV = true;
    recognition:SetTexture(self.CaptureInput);
    self.recognition    = recognition;

    local res = self.autotoucher:InitFaceBeautyRenderNode(scene,
                                                          self.hightlight_texture,
                                                          self.mask_texture,
                                                          self.teeth_lookup_texture,
                                                          self.eye_mask_texture,
                                                          layer_sequence);

    if res > 0 then
        ERROR("[FaceBeauty] InitFaceBeautyRenderNode return value: "..res);
        self.initialized = false;
    end
    else -- APP_IMO_VENUS
        WARNING("[FaceBeauty] imo product,  power off autotoucher")
    end -- APP_IMO_VENUS

    self.initialized = true;
end

function FaceBeauty:SetShow(is_show)
    self.is_show = is_show;
    self.autotoucher:SetFaceBeautyShow(is_show);
end


--[[
    This function should be called before Update(), it will
    1. disable all render first
    2. enable render and return true if has face or valid strength
]]
function FaceBeauty:PreUpdate(def)

    if not self.initialized then
        ERROR("[FaceBeauty] not initialized");
        return false;
    end

    -- reset state and disable all render first
    if not APP_IMO_VENUS then
    self.need_autotoucher_facebeauty = false;
    self.image_scaler:SetShow(false);
    self.vertical_box_blur:SetShow(false);
    self.horizontal_box_blur:SetShow(false);
    end -- APP_IMO_VENUS
    self.clear_whiten:SetShow(false);
    self.identity_render:SetShow(false);

    -- check if enabled
    if self.is_show == false then
        LOG("[FaceBeauty] don't show");
        return false;
    end

    if APP_IMO_VENUS then
        return true ;
    end -- APP_IMO_VENUS

    if _KRATOSEDITOR then

    else
        self.virtual_node.node:Update(def);
    end

    -- EraseRenderProperty the rendernodes created in AutoToucherFaceBeauty each time the function is called
    -- if autotoucher facebeauty items need to be updated, UpdateFaceBeautyRenderNode will reset the RP_SHOW property, else it will not be rendered.
    local res = self.autotoucher:PreUpdateFaceBeautyRenderNode(self.recognition, self.classify);

    self.need_autotoucher_facebeauty = (res == 0);

    if res ~= 0 then
        if self.preUpdateResult ~= res then
            ERROR("[FaceBeauty] PreUpdateFaceBeautyRenderNode return value: "..res);
        end
    end
    self.preUpdateResult = res;

    -- clear and whiten need to be updated even without face detected
    return true;
end


function FaceBeauty:Update(sequence, texture)
    local max_sequence = sequence - self.max_face_num;
    local texture_size = texture:GetSize();
    local ori_texture_for_face_beauty = texture;

    local fullUpdate = false ;

    if self.need_clear_or_whiten then
        self.clear_whiten:SetSequence(max_sequence - 4);
        self.clear_whiten:Update(texture_size:x(), texture_size:y());

        self.clear_whiten:SetParameter("UNIFORM_INPUT_TEXTURE", texture);
        self.clear_whiten:SetParameter("UNIFORM_CLEAR_INTENSITY", mathfunction.vector1(self.beauty_strengths[self.beauty_typeid.CLEAR]));
        self.clear_whiten:SetParameter("UNIFORM_WHITEN_INTENSITY", mathfunction.vector1(self.beauty_strengths[self.beauty_typeid.WHITEN]));

        ori_texture_for_face_beauty = self.clear_whiten:GetResultImage();

        self.identity_render:SetSequence(sequence);
        self.identity_render:GetNode():GetComponent(apolloengine.Node.CT_RENDER):SetRenderOrder(0);
        self.identity_render:SetShow(true);
        self.identity_render:SetParameter("UNIFORM_INPUT_TEXTURE", ori_texture_for_face_beauty);

        fullUpdate = true;
    else
        self.clear_whiten:SetShow(false);
        self.identity_render:SetShow(false);
        LOG("[FaceBeauty] no need to update clear and whiten");
    end

    if APP_IMO_VENUS then
        return fullUpdate
    end

    if not self.need_autotoucher_facebeauty then
        LOG("[FaceBeauty] no need to update autotoucher facebeauty items");
        return fullUpdate;
    end

    self.image_scaler:SetSequence(max_sequence - 3);
    self.vertical_box_blur:SetSequence(max_sequence - 2);
    self.horizontal_box_blur:SetSequence(max_sequence - 1);

    -- update quad render
    self:UpdateTextureSize(texture_size:x(), texture_size:y());
    self.image_scaler:SetParameter("UNIFORM_INPUT_TEXTURE", ori_texture_for_face_beauty);

    -- update face render
    local res = self.autotoucher:UpdateFaceBeautyRenderNode(ori_texture_for_face_beauty,
                                                            self.image_scaler:GetResultImage(),
                                                            self.horizontal_box_blur:GetResultImage(),
                                                            sequence);

    if res ~= 0 then
        if self.updateResult ~= res then
            ERROR("[FaceBeauty] UpdateFaceBeautyRenderNode return value: "..res);
        end
    end

    self.updateResult = res;

    return fullUpdate;
end

function FaceBeauty:SetBeautyStrength(beauty_type, strength)

    self.beauty_strengths[beauty_type] = strength;

    if self.beauty_strengths[self.beauty_typeid.CLEAR] > 0 or self.beauty_strengths[self.beauty_typeid.WHITEN] > 0 then
        self.need_clear_or_whiten = true;
    else
        self.need_clear_or_whiten = false;
    end

    if APP_IMO_VENUS then
       if beauty_type >= 0 and beauty_type <= 4 then
           WARNING("FaceBeauty, imo disable beauty_type "..tostring(beauty_type))
       end
       return true ;
    end

    if beauty_type >= 0 and beauty_type <= 4 then
        self.autotoucher:SetFaceBeautyStrengthWithType(strength, beauty_type);
    end

    return true;

end

function FaceBeauty:GetBeautyStrength(beauty_type)
    return self.beauty_strengths[beauty_type];
end

function FaceBeauty:GetMaskTexturePath()
  return self.mask_texture_path;
end

function FaceBeauty:SetMaskTexturePath(value)
  self.mask_texture_path = value;
  self.mask_texture      = self:LoadTexture(self.mask_texture_path);
end

function FaceBeauty:GetTeethLookupTexturePath()
  return self.teeth_lookup_texture_path;
end

function FaceBeauty:SetTeethLookupTexturePath(value)
  self.teeth_lookup_texture_path = value;
  self.teeth_lookup_texture      = self:LoadTexture(self.teeth_lookup_texture_path);
end

function FaceBeauty:GetEyeMaskTexturePath()
  return self.eye_mask_texture_path;
end

function FaceBeauty:SetEyeMaskTexturePath(value)
  self.eye_mask_texture_path = value;
  self.eye_mask_texture_texture = self:LoadTexture(self.eye_mask_texture_path);
end

function FaceBeauty:Destroy()
    if self.clear_whiten ~= nil then
        self.clear_whiten:Clear();
        self.clear_whiten = nil;
    end
    if self.identity_render ~= nil then
        self.identity_render:Destroy();
        self.identity_render = nil;
    end
    if self.image_scaler ~= nil then
        self.image_scaler:Clear();
        self.image_scaler = nil;
    end
    if self.vertical_box_blur ~= nil then
        self.vertical_box_blur:Clear();
        self.vertical_box_blur = nil;
    end
    if self.horizontal_box_blur ~= nil then
        self.horizontal_box_blur:Clear()
        self.horizontal_box_blur = nil;
    end

    if self.virtual_node ~= nil then
        self.virtual_node:Destroy();
        self.virtual_node = nil;
    end

    if self.autotoucher ~= nil then
        self.autotoucher:ReleaseFaceBeautyRenderNode();
        self.autotoucher = nil;
    end

    self.preUpdateResult = 0;
    self.updateResult    = 0;
    self.initialized     = false;
end

return FaceBeauty;