local Object       = require "classic"
local apolloengine = require "apolloengine"
local mathfunction = require "mathfunction"
local apollonode   = require "apolloutility.apollonode"
local venuscore    = require "venuscore"
local face_defined = require "facebeautyutils.defined"

--[[
    mesh render with exclusive camera.
]]
local MeshRender = Object:extend();

function MeshRender:new(render_name, render_sequence, mask_layer, material_path, width, height)

    self.node = apollonode.RenderNode();
    self.node:SetName(render_name);

    self.camera = self.node:CreateComponent(apolloengine.Node.CT_CAMERA);
    self.frame_width  = width;
    self.frame_height = height;
    self.initialized  = false;

    -- setup fbo
    local fbo_size     = mathfunction.vector2(self.frame_width, self.frame_height);
    self.render_target = apolloengine.RenderTargetEntity();
    self.render_target:PushMetadata(apolloengine.RenderTargetMetadata(apolloengine.RenderTargetEntity.RT_RENDER_TARGET_2D,
                                                                      apolloengine.RenderTargetEntity.ST_SWAP_UNIQUE,
                                                                      mathfunction.vector4(0, 0, self.frame_width, self.frame_height),
                                                                      fbo_size));
    self.fbo_texture = self.render_target:MakeTextureAttachment(apolloengine.RenderTargetEntity.TA_COLOR_0);
    self.fbo_texture:PushMetadata(apolloengine.TextureBufferMetadata(fbo_size,
                                                                     apolloengine.TextureEntity.TT_TEXTURE2D,
                                                                     apolloengine.TextureEntity.TU_STATIC,
                                                                     apolloengine.TextureEntity.PF_R8G8B8A8,
                                                                     1, false,
                                                                     apolloengine.TextureEntity.TW_CLAMP_TO_EDGE,
                                                                     apolloengine.TextureEntity.TW_CLAMP_TO_EDGE,
                                                                     apolloengine.TextureEntity.TF_LINEAR,
                                                                     apolloengine.TextureEntity.TF_LINEAR));
    self.render_target:CreateResource();

    -- setup camera
    local clear_color = mathfunction.Color(0.0, 0.0, 0.0, 1.0);
    self.camera:SetSequenceCulling(true);
    self.camera:SetLayerMaskNothing();
    self.camera:AddLayerMask(mask_layer);
    self.camera:SetSequence(render_sequence);
    self.camera:SetClearColor(clear_color);
    self.camera:FixedResolution(fbo_size);
    self.camera:AttachRenderTarget(self.render_target);
    self.camera:Deactivate();

    -- setup render node
    self.node:SetSequence(render_sequence);
    self.node:SetLayer(mask_layer);
    self.node:SetShow(false);
    self:InitVertexAndIndex();
    self.node:CreateResource(venuscore.IFileSystem:PathAssembly(material_path),
                             apolloengine.RenderComponent.RM_TRIANGLES,
                             self.vetex_stream,
                             self.index_stream);

    self.is_show = false; -- whether the render is to be shown
    self.initialized = true;
    LOG("[MeshRender] create success "..mask_layer)
end

function MeshRender:InitVertexAndIndex()
    self.vetex_stream = apolloengine.VertexStream();
    self.vetex_stream:SetVertexType(apolloengine.ShaderEntity.ATTRIBUTE_POSITION,
                                    apolloengine.VertexBufferEntity.DT_FLOAT,
                                    apolloengine.VertexBufferEntity.DT_FLOAT,
                                    2);
    self.vetex_stream:SetVertexType(apolloengine.ShaderEntity.ATTRIBUTE_COORDNATE0,
                                    apolloengine.VertexBufferEntity.DT_FLOAT,
                                    apolloengine.VertexBufferEntity.DT_FLOAT,
                                    2);
    self.vetex_stream:ReserveBuffer(face_defined.VERTEX_NUM);

    self.index_stream = apolloengine.IndicesStream();
    self.index_stream:SetIndicesType(apolloengine.IndicesBufferEntity.IT_UINT16);
    self.index_stream:ReserveBuffer(face_defined.TRIANGLE_NUM * 3);

    -- index
    local index_num = face_defined.TRIANGLE_NUM * 3;
    for i = 1, index_num do
        self.index_stream:PushIndicesData(face_defined.TRIANGLE_INDICES[i]);
    end

    -- vetex
    for i = 1, face_defined.VERTEX_NUM do
        local tex_coord = mathfunction.vector2(face_defined.STANDARD_FACE_MASK_TEX_COORD[2 * i - 1], face_defined.STANDARD_FACE_MASK_TEX_COORD[2 * i]);
        self.vetex_stream:PushVertexData(apolloengine.ShaderEntity.ATTRIBUTE_COORDNATE0, tex_coord);
        self.vetex_stream:PushVertexData(apolloengine.ShaderEntity.ATTRIBUTE_POSITION, mathfunction.vector2(0, 0));
    end
end

function MeshRender:SetSequence(sequence)
    if not self.initialized then
        LOG("[MeshRender] SetSequence fail");
        return;
    end
    self.camera:SetSequence(sequence);
    self.node:SetSequence(sequence);
end

function MeshRender:SetShow(is_show)
    if not self.initialized then
        LOG("[MeshRender] SetShow fail");
        return;
    end

    self.node:SetShow(is_show);

    if is_show then
        self.camera:Activate();
    else
        self.camera:Deactivate();
    end
    self.is_show = is_show;
end

function MeshRender:SetClearColor(color)
    if not self.initialized then
        LOG("[MeshRender] SetClearColor fail");
        return;
    end

    self.camera:SetClearColor(color);
    self.camera:Activate();
end

function MeshRender:UpdateVertex(keypoints)
    if not self.initialized then
        LOG("[MeshRender] UpdateVertex fail");
        return;
    end

    if keypoints ~= nil then
        local newpoint     =  mathfunction.vector2(0, 0);
        local vertex_index = self.vetex_stream:GetAttributeIndex(apolloengine.ShaderEntity.ATTRIBUTE_POSITION);

        for i = 1, face_defined.VERTEX_NUM do
            newpoint:Set(keypoints[i][1], keypoints[i][2]);
            self.vetex_stream:ChangeVertexDataWithAttributeFast(vertex_index, i, newpoint);
            --LOG("face point: " .. i .. ", " .. newpoint:x() .. ", " .. newpoint:y());
        end
        self.vetex_stream:SetReflushInterval(1, face_defined.VERTEX_NUM);
        self.node.render:ChangeVertexBuffer(self.vetex_stream);
    end
end

function MeshRender:Update(frame_width, frame_height)
    if not self.initialized then
        LOG("[MeshRender] Update fail");
        return false;
    end

    self:SetShow(true);

    if frame_width == self.frame_width and frame_height == self.frame_height then
        return false;
    end

    -- change fbo
    local fbo_size = mathfunction.vector2(frame_width, frame_height);
    self.camera:FixedResolution(fbo_size);

    self.frame_width  = frame_width;
    self.frame_height = frame_height;

    return true;
end

function MeshRender:SetParameter(key, value)
    if key == nil or value == nil then
        LOG("[MeshRender] SetParameter nil "..key);
        return;
    end

    if self.initialized then
        self.node:SetParameter(key, value);
    end
end

function MeshRender:GetResultImage()
    return self.fbo_texture;
end

function MeshRender:Clear()
    if self.node ~= nil then
        self.node:Destroy(); -- 清理node
        self.Camera = nil;
    end
end

return MeshRender;
