local apolloengine = require "apolloengine"
local apollonode   = require "apolloutility.apollonode"
local mathfunction = require "mathfunction"
local venuscore    = require "venuscore"
local cv           = require "computervisionfunction"
local quad_render  = require "facebeautyutils.face_beauty_quad_render"
local mesh_render  = require "facebeautyutils.mesh_render"
local face_defined = require "facebeautyutils.defined"

--[[
    The same as gpu skin smooth proposal 6 in bigolive.
]]
local SkinSmooth = {};

function SkinSmooth:Create(layer_sequence, scene)
    self.is_show = true;
    self.initialized = false;
    self.enable_test = false;
    self.begin_time = 0;

    self:InitializeConfig();
    self:InitializeRender(layer_sequence, scene);
    self.initialized = true;

    -- enable test to show intermediate render result
    -- self:EnableTest(layer_sequence, self.facemask_render);
end

function SkinSmooth:InitializeConfig()
    self.facemask_texture_path = "docs:skinsmooth/resource/face_mask301.png";
    self.skin_smooth_strength = 0.0;
    self.face_num = 0;

    -- image scaled factor
    self.scaled_factor1 = 0.5; -- for blur image
    self.scaled_factor2 = 0.25; -- for var skin detection and facemask image

    -- selective blur uniform params
    self.max_distance = 0.7078; -- (60 * 3 + 0.5) / 255 = 0.7078

    -- highpass skin detection uniform params
    -- 1.0f / ((m_brightest_pix_diff / 255.0) * (m_brightest_pix_diff / 255.0)) (m_brightest_pix_diff = 20)
    -- for original var function use 50.f, for linear inverse var function use 162.5625f
    self.magnify_ratio = 50.0;
    -- if no face detected, use the default mean skin value for skin detection
    -- 601full meanYUV={135, 110, 150}  -->  meanRGB={0.6531, 0.4900, 0.4078}
    self.skin_meanRGB_default = {0.6531, 0.4900, 0.4078};
    -- set skin mean rgb value. if the value is not set, use the default value
    self.skin_meanRGB = self.skin_meanRGB_default;
    -- max(|R-meanR|),  max(|G-meanG|), max(|B-meanB|), min(R-G), min(R-B)
    self.skin_threshold = {0.3824, 0.3824, 0.3824, 0.0294, 0.0};

    -- skin smooth uniform params
    self.variance_tolerance = 0.15;
end

function SkinSmooth:InitializeRender(layer_sequence, scene)
    -- load texture
    self.facemask_texture = self:LoadTexture(self.facemask_texture_path);

    local initial_width  = 360;
    local initial_height = 640;

    -- init render
    self.facemask_render = mesh_render("facemask", layer_sequence - 6, "facemask_layer", "docs:skinsmooth/material/facemask.material", initial_width, initial_height);
    self.vertical_selective_blur_by_facemask_render = quad_render("vertical_selective_blur", layer_sequence - 5, "vertical_selective_blur_layer", "docs:skinsmooth/material/selective_blur_by_facemask_uniform_sample.material", initial_width, initial_height);
    self.horizontal_selective_blur_by_facemask_render = quad_render("horizontal_selective_blur", layer_sequence - 4, "horizontal_selective_blur_layer", "docs:skinsmooth/material/selective_blur_by_facemask_uniform_sample.material", initial_width, initial_height);
    self.highpass_skin_detection_render = quad_render("highpass_skin", layer_sequence - 3, "highpass_skin_layer", "docs:skinsmooth/material/highpass_skin_detection.material", initial_width, initial_height);
    self.vertical_var_skin_box_blur_uniform_render = quad_render("vertical_box_blur", layer_sequence - 2, "vertical_box_blur_layer", "docs:skinsmooth/material/boxblur_uniform_sample.material", initial_width, initial_height);
    self.horizontal_var_skin_box_blur_uniform_render = quad_render("horizontal_box_blur", layer_sequence - 1, "horizontal_box_blur_layer", "docs:skinsmooth/material/boxblur_uniform_sample.material", initial_width, initial_height);
    self.skin_smooth_blend_render = self:InitQuadNode(layer_sequence, "docs:skinsmooth/material/skin_smooth_blend.material");

    self.facemask_render:SetParameter("UNIFORM_INPUT_TEXTURE", self.facemask_texture);
    self.vertical_selective_blur_by_facemask_render:SetParameter("UNIFORM_FACEMASK_TEXTURE", self.facemask_render:GetResultImage());
    self.vertical_selective_blur_by_facemask_render:SetParameter("UNIFORM_MAX_DISTANCE", mathfunction.vector1(self.max_distance));
    self.horizontal_selective_blur_by_facemask_render:SetParameter("UNIFORM_INPUT_TEXTURE", self.vertical_selective_blur_by_facemask_render:GetResultImage());
    self.horizontal_selective_blur_by_facemask_render:SetParameter("UNIFORM_FACEMASK_TEXTURE", self.facemask_render:GetResultImage());
    self.horizontal_selective_blur_by_facemask_render:SetParameter("UNIFORM_MAX_DISTANCE", mathfunction.vector1(self.max_distance));
    self.highpass_skin_detection_render:SetParameter("UNIFORM_BLUR_TEXTURE", self.horizontal_selective_blur_by_facemask_render:GetResultImage());
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MAGNIFY_RATIO", mathfunction.vector1(self.magnify_ratio));
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MAX_RESR", mathfunction.vector1(self.skin_threshold[1]));
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MAX_RESG", mathfunction.vector1(self.skin_threshold[2]));
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MAX_RESB", mathfunction.vector1(self.skin_threshold[3]));
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MIN_RSG", mathfunction.vector1(self.skin_threshold[4]));
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MIN_RSB", mathfunction.vector1(self.skin_threshold[5]));
    self.vertical_var_skin_box_blur_uniform_render:SetParameter("UNIFORM_INPUT_TEXTURE", self.highpass_skin_detection_render:GetResultImage());
    self.horizontal_var_skin_box_blur_uniform_render:SetParameter("UNIFORM_INPUT_TEXTURE", self.vertical_var_skin_box_blur_uniform_render:GetResultImage());
    self.skin_smooth_blend_render:SetParameter("UNIFORM_BLUR_TEXTURE", self.horizontal_selective_blur_by_facemask_render:GetResultImage());
    self.skin_smooth_blend_render:SetParameter("UNIFORM_VAR_SKIN_MASK_TEXTURE", self.horizontal_var_skin_box_blur_uniform_render:GetResultImage());
    self.skin_smooth_blend_render:SetParameter("UNIFORM_FACEMASK_TEXTURE", self.facemask_render:GetResultImage());
    self.skin_smooth_blend_render:SetParameter("UNIFORM_VARIANCE_TOLERANCE", mathfunction.vector1(self.variance_tolerance));

    self.CaptureInput = apolloengine.TextureEntity();
    self.CaptureInput:PushMetadata(apolloengine.TextureReferenceMetadata(apolloengine.DeviceResource.DEVICE_CAPTURE));
    self.CaptureInput:CreateResource();

    self.virtual_node = apollonode.VirtualNode();

    local recognition   = self.virtual_node:CreateComponent(apolloengine.Node.CT_CV_RECOGNITION);
    recognition.Enable  = true;
    recognition.Mode    = cv.IVisionComponent.VIDEO;
    recognition.Type    = cv.RecognitionComponent.cvFace + cv.RecognitionComponent.cvForeheadDetection;
    recognition.ShareCV = true;
    recognition:SetTexture(self.CaptureInput);
    self.recognition    = recognition;
end

function SkinSmooth:LoadTexture(texture_path)
    local texture = apolloengine.TextureEntity();
    texture:PushMetadata(apolloengine.TextureFileMetadata(
                                   apolloengine.TextureEntity.TU_STATIC,
                                   apolloengine.TextureEntity.PF_AUTO,
                                   1, false,
                                   apolloengine.TextureEntity.TW_REPEAT,
                                   apolloengine.TextureEntity.TW_REPEAT,
                                   apolloengine.TextureEntity.TF_LINEAR,
                                   apolloengine.TextureEntity.TF_LINEAR,
                                   venuscore.IFileSystem:PathAssembly(texture_path)));
    texture:SetJobType(venuscore.IJob.JT_SYNCHRONOUS);
    texture:CreateResource();
    return texture;
end

function SkinSmooth:InitQuadNode(seq, material_path)
    local render = apollonode.QuadNode();
    render:CreateResource(material_path, true);
    render:SetSequence(seq);
    render:GetNode():GetComponent(apolloengine.Node.CT_RENDER):SetRenderOrder(1);
    render:SetCull(false);
    render:SetShow(false);
    render:SetName("identity render");
    -- render:GetNode().Active = false;

    return render;
end

function SkinSmooth:SetSkinSmoothStrength(strength)
    self.skin_smooth_strength = strength;
    return true;
end

function SkinSmooth:SetMeanSkinRGBColor(mean_r, mean_g, mean_b)
    if not self.initialized then
        ERROR("[SkinSmooth] not initialized");
        return false;
    end
    self.skin_meanRGB = {mean_r, mean_g, mean_b};
    return true;
end

function SkinSmooth:SetShow(is_show)
    self.is_show = is_show;
end

function SkinSmooth:IsWorking()
    if not self.initialized then
        ERROR("[SkinSmooth] not initialized");
        return false;
    end

    -- check if enabled
    if self.skin_smooth_strength < 0.001 or self.is_show == false then
        LOG("[SkinSmooth] don't show");
        return false;
    end

    return true;
end

function SkinSmooth:Disable()
    if not self.initialized then
        ERROR("[SkinSmooth] not initialized");
        return;
    end

    -- reset state and disable all render
    self.face_num = 0;
    self.facemask_render:SetShow(false);
    self.vertical_selective_blur_by_facemask_render:SetShow(false);
    self.horizontal_selective_blur_by_facemask_render:SetShow(false);
    self.highpass_skin_detection_render:SetShow(false);
    self.vertical_var_skin_box_blur_uniform_render:SetShow(false);
    self.horizontal_var_skin_box_blur_uniform_render:SetShow(false);
    self.skin_smooth_blend_render:SetShow(false);
    if self.enable_test then
        self.identity_render:SetShow(false); -- test render result
    end
end

function SkinSmooth:UpdateKeypoints(def)
    if _KRATOSEDITOR then

    else
        self.virtual_node.node:Update(def);
    end

    local results = self.recognition:GetResult();
    local detect_size = self.recognition:GetDetectSize();
    if results == nil or detect_size == nil then
        self.face_num = 0;
        LOG("[SkinSmooth] no face detected");
        return true;
    end
    local face_results = results[cv.RecognitionComponent.cvFace];
    local forehead_results = results[cv.RecognitionComponent.cvForeheadDetection];
    local face_keypoints = nil;
    local forehead_keypoints = nil;

    if face_results == nil or forehead_results == nil then
        self.face_num = 0;
    else
        local n_faces = #face_results;
        local n_foreheads = #forehead_results;
        if n_faces == 0 or n_faces ~= n_foreheads then
            self.face_num = 0;
        else
            face_keypoints = face_results[1]; -- only process one face
            forehead_keypoints = forehead_results[1];
            if face_keypoints ~= nil and forehead_keypoints ~= nil then
                self.face_num = n_faces;
            else
                self.face_num = 0;
            end
        end
    end

    -- LOG("[SkinSmooth] face num: "..self.face_num);

    if self.face_num > 0 then
        local size_inv = {2.0 / detect_size:x(), 2.0 / detect_size:y()};
        local normalized_keypoints = {};

        -- 1 -> 104: 106 face keypoints - 2 repeated face keypoints
        for j = 1, face_defined.FACE_LANDMARK_NUM do
            local point = {mx=0, my=0};
            point.mx = face_keypoints[j * 2 - 1];
            point.my = face_keypoints[j * 2];
            normalized_keypoints[j] = {point.mx * size_inv[1] - 1.0, 1.0 - point.my * size_inv[2]};
        end
        local total_keypoints = face_defined.FACE_LANDMARK_NUM;

        -- 105 -> 127: 23 forehead keypoints from right to left
        for j = 1, face_defined.FOREHEAD_LANDMARK_NUM do
            local point = {mx=0, my=0};
            point.mx = forehead_keypoints[j * 2 - 1];
            point.my = forehead_keypoints[j * 2];
            normalized_keypoints[total_keypoints + j] = {point.mx * size_inv[1] - 1.0, 1.0 - point.my * size_inv[2]};
        end
        total_keypoints = total_keypoints + face_defined.FOREHEAD_LANDMARK_NUM;

        -- 128 -> 160: 33 outer face keypoints
        for j = 1, 10 do
            local delta = 0.5 + (j - 1) / 40.0;
            normalized_keypoints[total_keypoints + j] = {(1.0 + delta) * normalized_keypoints[j][1] - delta * normalized_keypoints[47][1],
                                                         (1.0 + delta) * normalized_keypoints[j][2] - delta * normalized_keypoints[47][2]};
        end
        for j = 11, 17 do
            local delta = 0.8 + (j - 11) / 20.0;
            normalized_keypoints[total_keypoints + j] = {(1.0 + delta) * normalized_keypoints[j][1] - delta * normalized_keypoints[47][1],
                                                         (1.0 + delta) * normalized_keypoints[j][2] - delta * normalized_keypoints[47][2]};
        end
        for j = 18, 23 do
            local delta = 0.8 + (24 - j) / 20.0;
            normalized_keypoints[total_keypoints + j] = {(1.0 + delta) * normalized_keypoints[j][1] - delta * normalized_keypoints[47][1],
                                                         (1.0 + delta) * normalized_keypoints[j][2] - delta * normalized_keypoints[47][2]};
        end
        for j = 24, 33 do
            local delta = 0.5 + (34 - j) / 40.0;
            normalized_keypoints[total_keypoints + j] = {(1.0 + delta) * normalized_keypoints[j][1] - delta * normalized_keypoints[47][1],
                                                         (1.0 + delta) * normalized_keypoints[j][2] - delta * normalized_keypoints[47][2]};
        end
        total_keypoints = total_keypoints + face_defined.OUTER_FACE_LANDMARK_NUM;

        -- 161 -> 183: 23 outer forehead keypoint
        for j = 1, face_defined.OUTER_FOREHEAD_LANDMARK_NUM do
            local delta = 0.5;
            normalized_keypoints[total_keypoints + j] = {(1.0 + delta) * normalized_keypoints[face_defined.FACE_LANDMARK_NUM + j][1] - delta * normalized_keypoints[47][1],
                                                         (1.0 + delta) * normalized_keypoints[face_defined.FACE_LANDMARK_NUM + j][2] - delta * normalized_keypoints[47][2]};
        end
        total_keypoints = total_keypoints + face_defined.OUTER_FOREHEAD_LANDMARK_NUM;

        self.facemask_render:UpdateVertex(normalized_keypoints);
    end

    return true;
end

function SkinSmooth:GetMeanSkinRGBFromRecognition()
    if self.face_num > 0  then
        local meanRGB = self.recognition:GetMeanSkinRGB();
        if meanRGB[1] == 0 or meanRGB[1] == 0 or meanRGB[2] == 0 then
            self.skin_meanRGB = self.skin_meanRGB_default;
        else
            self.skin_meanRGB = {meanRGB[1] / 255.0, meanRGB[2] / 255.0, meanRGB[3] / 255.0};
        end
    else
        self.skin_meanRGB = self.skin_meanRGB_default;
    end
    -- LOG("meanRGB: "..self.skin_meanRGB[1].." "..self.skin_meanRGB[2].." "..self.skin_meanRGB[3]);
end

function SkinSmooth:UpdateMeanSkinRGB()
    local now = venuscore.ITimerSystem:GetTimevalue();
    local interval = now - self.begin_time;
    if interval > 1 then
        self:GetMeanSkinRGBFromRecognition();
        self.begin_time = now;
    end
end

function SkinSmooth:Update(def, layer_sequence, texture)
    self:UpdateKeypoints(def);
    self:UpdateMeanSkinRGB();

    local texture_size = texture:GetSize();

    -- clear facemask buffer first, the binded texture is set to the specified clear_color even when facemask_render is not shown
    local clear_color = mathfunction.Color(0.0, 0.0, 0.0, self.face_num > 1 and 0.8 or 0.2);
    self.facemask_render:SetClearColor(clear_color);
    if self.face_num > 0 then
        self.facemask_render:SetSequence(layer_sequence - 6);
        self.facemask_render:Update(texture_size:x() * self.scaled_factor2, texture_size:y() * self.scaled_factor2);
        self.facemask_render:SetParameter("UNIFORM_FACE_NUM", mathfunction.vector1(self.face_num));
    end

    self.vertical_selective_blur_by_facemask_render:SetSequence(layer_sequence - 5);
    self.horizontal_selective_blur_by_facemask_render:SetSequence(layer_sequence - 4);
    self.highpass_skin_detection_render:SetSequence(layer_sequence - 3);
    self.vertical_var_skin_box_blur_uniform_render:SetSequence(layer_sequence - 2);
    self.horizontal_var_skin_box_blur_uniform_render:SetSequence(layer_sequence - 1);
    self.skin_smooth_blend_render:SetSequence(layer_sequence);
    self.skin_smooth_blend_render:GetNode():GetComponent(apolloengine.Node.CT_RENDER):SetRenderOrder(1);
    self.skin_smooth_blend_render:SetShow(true);

    self:UpdateTextureSize(texture_size:x(), texture_size:y());

    self.vertical_selective_blur_by_facemask_render:SetParameter("UNIFORM_INPUT_TEXTURE", texture);
    self.highpass_skin_detection_render:SetParameter("UNIFORM_INPUT_TEXTURE", texture);
    self.highpass_skin_detection_render:SetParameter("UNIFORM_MEAN_COLOR", mathfunction.vector3(self.skin_meanRGB[1], self.skin_meanRGB[2], self.skin_meanRGB[3]));
    self.skin_smooth_blend_render:SetParameter("UNIFORM_INPUT_TEXTURE", texture);
    self.skin_smooth_blend_render:SetParameter("UNIFORM_SMOOTH_INTENSITY", mathfunction.vector1(self.skin_smooth_strength));
    self.skin_smooth_blend_render:SetParameter("UNIFORM_FACE_NUM", mathfunction.vector1(self.face_num));

    -- test render result
    if self.enable_test then
        self.identity_render:SetSequence(layer_sequence);
        self.identity_render:GetNode():GetComponent(apolloengine.Node.CT_RENDER):SetRenderOrder(2);
        self.identity_render:SetShow(true);
        self.identity_render:SetParameter("UNIFORM_INPUT_TEXTURE", self.image_to_show);
    end
end

function SkinSmooth:UpdateTextureSize(frame_width, frame_height)

    local scaled_width1  = frame_width * self.scaled_factor1;
    local scaled_height1 = frame_height * self.scaled_factor1;
    local scaled_width2  = frame_width * self.scaled_factor2;
    local scaled_height2 = frame_height * self.scaled_factor2;

    self.highpass_skin_detection_render:Update(scaled_width2, scaled_height2);

    local is_scaled1 = self.vertical_selective_blur_by_facemask_render:Update(scaled_width1, scaled_height1);
    local is_scaled2 = self.horizontal_selective_blur_by_facemask_render:Update(scaled_width1, scaled_height1);
    local is_scaled3 = self.vertical_var_skin_box_blur_uniform_render:Update(scaled_width2, scaled_height2);
    local is_scaled4 = self.horizontal_var_skin_box_blur_uniform_render:Update(scaled_width2, scaled_height2);

    -- update
    if is_scaled1 or is_scaled2 or is_scaled3 or is_scaled4 then
        self.vertical_selective_blur_by_facemask_render:SetParameter("UNIFORM_PIXEL_BASE_OFFSET", mathfunction.vector2(0.0, 1.5 / scaled_height1));
        self.vertical_selective_blur_by_facemask_render:SetParameter("UNIFORM_PIXEL_STRIDE", mathfunction.vector2(0.0, 2.0 / scaled_height1));
        self.horizontal_selective_blur_by_facemask_render:SetParameter("UNIFORM_PIXEL_BASE_OFFSET", mathfunction.vector2(1.5 / scaled_width1, 0.0));
        self.horizontal_selective_blur_by_facemask_render:SetParameter("UNIFORM_PIXEL_STRIDE", mathfunction.vector2(2.0 / scaled_width1, 0.0));

        self.vertical_var_skin_box_blur_uniform_render:SetParameter("UNIFORM_PIXEL_BASE_OFFSET", mathfunction.vector2(0.0, 1.5 / scaled_height2));
        self.vertical_var_skin_box_blur_uniform_render:SetParameter("UNIFORM_PIXEL_STRIDE", mathfunction.vector2(0.0, 2.0 / scaled_height2));
        self.horizontal_var_skin_box_blur_uniform_render:SetParameter("UNIFORM_PIXEL_BASE_OFFSET", mathfunction.vector2(1.5 / scaled_width2, 0.0));
        self.horizontal_var_skin_box_blur_uniform_render:SetParameter("UNIFORM_PIXEL_STRIDE", mathfunction.vector2(2.0 / scaled_width2, 0.0));
    end

  end

-- test render result
function SkinSmooth:EnableTest(layer_sequence, render)
    LOG("[SkinSmooth] test enable!");

    self.enable_test = true;
    self.identity_render = self:InitQuadNode(layer_sequence, "docs:facebeauty/material/identity.material");
    self.image_to_show = render:GetResultImage();
end

function SkinSmooth:Destroy()
    if self.facemask_render ~= nil then
        self.facemask_render:Clear();
        self.facemask_render = nil;
    end
    if self.vertical_selective_blur_by_facemask_render ~= nil then
        self.vertical_selective_blur_by_facemask_render:Clear();
        self.vertical_selective_blur_by_facemask_render = nil;
    end
    if self.horizontal_selective_blur_by_facemask_render ~= nil then
        self.horizontal_selective_blur_by_facemask_render:Clear();
        self.horizontal_selective_blur_by_facemask_render = nil;
    end
    if self.highpass_skin_detection_render ~= nil then
        self.highpass_skin_detection_render:Clear();
        self.highpass_skin_detection_render = nil;
    end
    if self.vertical_var_skin_box_blur_uniform_render ~= nil then
        self.vertical_var_skin_box_blur_uniform_render:Clear();
        self.vertical_var_skin_box_blur_uniform_render = nil;
    end
    if self.horizontal_var_skin_box_blur_uniform_render ~= nil then
        self.horizontal_var_skin_box_blur_uniform_render:Clear();
        self.horizontal_var_skin_box_blur_uniform_render = nil;
    end
    if self.skin_smooth_blend_render ~= nil then
        self.skin_smooth_blend_render:Destroy();
        self.skin_smooth_blend_render = nil;
    end
    if self.identity_render ~= nil then
        self.identity_render:Destroy();
        self.identity_render = nil;
    end
    if self.virtual_node ~= nil then
        self.virtual_node:Destroy();
        self.virtual_node = nil;
    end

    self.initialized = false;
end

return SkinSmooth;