local apolloengine = require "apolloengine"
local mathfunction = require "mathfunction"
local venuscore = require "venuscore"

local dither = {}

local MBIG =  2147483647;
local MSEED = 161803398;
local MinValue = -21474836482;

function dither:Random(Seed)
  local ii;
  local mj, mk;

  --Initialize our Seed array.
  --This algorithm comes from Numerical Recipes in C (2nd Ed.)
  local subtraction = (Seed == MinValue) and MBIG or math.abs(Seed);
  mj = MSEED - subtraction;
  self.SeedArray[56]=mj;

  mk=1;
  for  i=1, 54 do --Apparently the range [1..55] is special (Knuth) and so we're wasting the 0'th position.
    ii = (21*i)%55;
    self.SeedArray[ii + 1]=mk;
    mk = mj - mk;
    if mk<0 then
      mk = mk +MBIG
    end
    mj=self.SeedArray[ii + 1];
  end

  for k=1, 4 do
    for i=1, 55 do
      self.SeedArray[i] = self.SeedArray[i] - self.SeedArray[1+(i+30)%55];
      if self.SeedArray[i]<0 then
        self.SeedArray[i] = self.SeedArray[i] + MBIG;
      end
    end
  end
  self.inext=0;
  self.inextp = 21;
  Seed = 1;
end

function dither:InternalSample()
  local retVal;
  local locINext = self.inext;
  local locINextp = self.inextp;

  locINext = locINext + 1;
  if (locINext >=57) then
    locINext=1;
  end

  locINextp = locINextp + 1;
  if (locINextp>= 57) then
    locINextp = 1;
  end

  retVal = self.SeedArray[locINext]-self.SeedArray[locINextp];

  if (retVal == MBIG) then
    retVal = retVal - 1;
  end
  if (retVal<0) then
    retVal= retVal + MBIG;
  end

  self.SeedArray[locINext]=retVal;

  self.inext = locINext;
  self.inextp = locINextp;

  return retVal;
end

function dither:Sample()
  --Including this division at the end gives us significantly improved
  --random number distribution.
  return (self:InternalSample()*(1.0/MBIG));
end

function dither:Initialize(host, size)

  self.Queue = 310;

  self.inext = 0;
  self.inextp = 0;
  self.SeedArray = {};  --56

  for i = 1, 56 do
    self.SeedArray[i] = 0;
  end

  self:Random(1234);

  self.noiseTexEntity = apolloengine.TextureEntity();
  self.noiseTexEntity:PushMetadata(apolloengine.TextureFileMetadata(
          apolloengine.TextureEntity.TU_STATIC,
          apolloengine.TextureEntity.PF_AUTO,1, false,
          apolloengine.TextureEntity.TW_CLAMP_TO_EDGE,
          apolloengine.TextureEntity.TW_CLAMP_TO_EDGE,
          apolloengine.TextureEntity.TF_NEAREST,
          apolloengine.TextureEntity.TF_NEAREST,
          "comm:documents/texture/material/black.png"));
  self.noiseTexEntity:SetJobType(venuscore.IJob.JT_SYNCHRONOUS);
  self.noiseTexEntity:CreateResource();

  self.LastLutTexEntity = self.LutTexEntity;
  self.LastNoiseTexEntity = self.noiseTexEntity;

  self._DitheringTex = apolloengine.IMaterialSystem:NewParameterSlot(
          apolloengine.ShaderEntity.UNIFORM,
          "_DitheringTex");

  self._MainTex = apolloengine.IMaterialSystem:NewParameterSlot(
          apolloengine.ShaderEntity.UNIFORM,
          "_MainTex");


  local pathout = venuscore.IFileSystem:PathAssembly("comm:documents/material/unity_bloom/FinalPass.material");
  self.finalMA     = host:CreateMaterial( pathout );
  self.finalRO     = host:CreateRenderObject();

  self.finalMA:SetParameter(self._DitheringTex, self.noiseTexEntity);

  self._Dithering_Coords = apolloengine.IMaterialSystem:NewParameterSlot(apolloengine.ShaderEntity.UNIFORM, "_Dithering_Coords");
  self.DitheringCoords = mathfunction.vector4(0.0,0.0,0.0,0.0);

  self.fbosize  = mathfunction.vector2(size:x(), size:y());

  self.noistTexSize = mathfunction.vector2(64, 64);

  host:RegisterScriptParameter(self, "noiseTexEntity");
  return self.Queue;
end

function dither:Resizeview(size)
  self.fbosize = mathfunction.vector2(size:x(), size:y());
end

function dither:Process(context, Original, Scene, Output)

  local rndOffsetX = self:Sample();
  local rndOffsetY = self:Sample();

  self.noistTexSize = self.noiseTexEntity:GetSize();
  local Dithering_Coords = mathfunction.vector4(self.fbosize.mx / self.noistTexSize.mx, self.fbosize.my / self.noistTexSize.my, rndOffsetX, rndOffsetY);

  context:BeginRenderPass(Output, apolloengine.RenderTargetEntity.CF_COLOR);
  self.finalMA:SetParameter(self._MainTex,
          Scene:GetAttachment( apolloengine.RenderTargetEntity.TA_COLOR_0 ) );
  if self.noiseTexEntity ~= self.LastNoiseLutTexEntity then
    self.finalMA:SetParameter(self._DitheringTex, self.noiseTexEntity);
    self.LastNoiseLutTexEntity = self.noiseTexEntity;
  end
  self.finalMA:SetParameter( self._Dithering_Coords, Dithering_Coords);
  context:Draw(self.finalRO, self.finalMA);
  context:EndRenderPass();

end

return dither;